unit PascalUnit;
{* This unit contains code, which can be used to recognize syntax and
   semantics of Pascal unit. It is intended to be used as utility part
   of XCL:
   |<br>
   1. To generate this help on base of comments, placed in interface
      part of XCL units.
   |<br>
   2. To manage source of XCL unit in XCL Visual Wizard, embedded into
      Delphi IDE environment.
   |<br>
   Certainly, this unit can be used for other purposes (may be, to
   create another Pascal compiler ? :)
}

interface

uses
  Windows, Messages, Kol;

type
  TEntryType = ( etField, etMethod, etProperty,
                 etType, etConst );
  {* Available types of declaration entries (standalone or parts of
     class or record definition). }
  TEntryTypeSet = Set of TEntryType;
  {* Set of available entries type - to pass it as a parameter for
     methods, searching certain entries. }
  TSectionType = ( stPrivate, stProtected, stPublic );
  {* Available types of declaration sections of class ('published' not
     included here - because XCL never use it). }
  TSectionTypeSet = Set of TSectionType;
  {* Set of available section types. Just to pass it as a parameter
     for methods, searching certain declaration entries. }

  PEntryRecordClass = ^TEntryRecordClass;
  TEntryRecordClass = object( TObj )
  private
    fIsEntry: Boolean;
  protected
  public
    property IsEntry: Boolean read fIsEntry;
  end;

  PPascalUnit = ^TPascalUnit;
  PRecordClass = ^TRecordClass;

  PEntry = ^TEntry;
  TEntry = object( TEntryRecordClass )
  {* Incapsulation of a single declaration entry. }
  private
    FOwner : PRecordClass;
    FUnit : PPascalUnit;
    FName : String;
    FType : TEntryType;
    FSection : TSectionType;
    FStartPos : TPoint;
    FEndPos : TPoint;
    FBeginImplement : TPoint;
    FBeginBody : TPoint;
    FEndBody : TPoint;
    function GetBeginBody: TPoint;
    function GetEndBody: TPoint;
  protected
  public
    //constructor Create( Owner : ZRecordClass; AUnit : ZPascalUnit );
    {* Constructor. If Owner = nil, declaration is standalone (in interface
       section). }
    property Name : String read FName;
    {* Name of declared element (type, variable, constant, etc.) }
    property EntryType : TEntryType read FType;
    {* Type of entry. }
    property Section : TSectionType read FSection;
    {* Section. For standalone declarations, stPublic is used. }
    property StartPos : TPoint read FStartPos;
    {* Coordinates of starting character of the declaration entry. }
    property EndPos : TPoint read FEndPos;
    {* Coordinates of a character, following the last one in declaration. }
    property BeginBody : TPoint read GetBeginBody;
    {* Coordinates of starting character of method implementation
       (found in implementation section) }
    property EndBody : TPoint read GetEndBody;
    {* Coordinates of a character, following the last one in method body. }
    function IsBodyEmpty : Boolean;
    {* Returns True, if there are no significiant characters between
       'begin' and 'end' tokens of method implementation. Always False, if
       entry is not a method declaration. }
  end;

  TRecordClass = object( TEntryRecordClass )
  {* Incapsulation of declarations set for class or record. }
  private
    FUnit : PPascalUnit;
    FEntries : PList;
    FName, FAncestorName : String;
    FDefStart, FDefEnd : TPoint;
    fDefBodyStart: TPoint;
    procedure BuildDefinition;
    procedure ClearEntries;
    function GetEntries(Idx: Integer): PEntry;
    function GetEntryCount: Integer;
  protected
  public
    //constructor Create( Source : PPascalUnit; FStart : TPoint );
    {* Constructor. }
    destructor Destroy; virtual;
    {* Destructor. }
    property EntryCount : Integer read GetEntryCount;
    {* Number of entries in class/record declaration. }
    property Entries[ Idx : Integer ] : PEntry read GetEntries;
    {* Array of entries in class/record declaration. }
    function FindEntry( EntryTypes : TEntryTypeSet; Sections : TSectionTypeSet;
                        const AName : String ) : PEntry;
    {* Returns entry with given Name (without case sensitivity) searching
       it among all entries of given type and given sections. }
    procedure SearchMethodImplementations;
    {* Call this method to find implementations for all method declarations,
       found in class definition. }
    property DefStart : TPoint read FDefStart;
    {* Coordinates of first character in class declaration. }
    property DefBodyStart: TPoint read fDefBodyStart;
    {* Coordinates of the first character following class/object/record header. }
    property DefEnd : TPoint read FDefEnd;
    {* Coordinates of a character, following to the last symbol in
       class/record definition. }
    property Name : String read FName;
    {* Name of class/record type. }
    property Ancestor : String read FAncestorName;
    {* Name of class ancestor. }
  end;

  TPascalUnit = object( TObj )
  {* Incapsulation of Delphi Pascal unit. }
  private
    FUnitName: String;
    FUnitPath: String;
    FText: PStrList;
    FCleanText: PStrList;
    FImplementationPos : TPoint;
    FUsesClause1Pos, FUsesClause2Pos : TPoint;
    FUsesClause1End : TPoint;
    FStartTokenPos : TPoint;
    FDecls : PList;
    FUses1Entries, FUses2Entries : PStrList;
    procedure SetUnitPath(const Value: String);
    function GetCleanText(Idx: Integer): String;
    procedure ScanUsesEntries( WhatUses: PStrList; var Pos : TPoint );
    function GetDecls(Idx: Integer): PObj;
    function GetDeclsCount: Integer;
    procedure Load;
    procedure ClearDeclsAndGlobals;
  protected
  public
    //constructor Create;
    {* Constructor. }
    destructor Destroy; virtual;
    {* Destructor. }
    procedure Clear;
    {* Call this method to clear ZPascalUnit object. }
    property UnitPath : String read FUnitPath write SetUnitPath;
    {* Path to unit source on disk (if available). }
    property UnitName : String read FUnitName;
    {* Name of unit. }
    procedure BuildDefinitions;
    {* Call this method to create all definitions. }
    property Text : PStrList read FText;
    {* Access to source of a unit. }
    property CleanText[ Idx : Integer ] : String read GetCleanText;
    {* Source of the unit, where all comments are replaced with spaces (to
       simplify scanning of a source). }
    property DeclCount : Integer read GetDeclsCount;
    {* Number of declarations in interface section. }
    property Decls[ Idx : Integer ]: PObj read GetDecls;
    {* Array of available declarations. }

    function ScanToToken( const Token : String; var Pos : TPoint;
                          BeforeImplementation, IgnoreData, NoMethods : Boolean ) : Boolean;
    {* Call this method, when You are sure, that given token *must* finish
       following to Pos sequence of tokens - to skip all ones before. }
    function ScanToEnd( var Pos: TPoint ): Boolean;
    {* Call this function to skip all to 'end' token, taking into account internal begin-end,
       record-end, class-end, object-end bracket pairs. }
    function Scan( var Pos : TPoint ) : String;
    {* Call this method to scan a single token (identifier, number, string constant
       or symbol). }
    procedure SkipToBracket( var Pos : TPoint );
    {* Call this method to skip all tokens in round brackets, taking into
       consideration possible pairs of round brackets. Useful to skip
       sub-expressions, or constant initialization blocks. }
    procedure SkipToSqrBracket( var Pos : TPoint );
    {* Call this method to skip all tokens in square brackets, taking into
       consideration possible pairs of square brackets. Useful to skip
       indexes or array definitions. }
    procedure SkipTypeDefinitions( var Pos : TPoint );
    {* Skip all type definitions. Usually should be called after token 'TYPE'. }
    procedure SkipTypeDefinition( var Pos : TPoint );
    {* Skip a single type definition. }
    procedure SkipDataDefinitions( var Pos : TPoint );
    {* Skip all data definitions (after token 'VAR'). }
    function SkipDataDefinition( var Pos : TPoint ) : Boolean;
    {* Skip a single data definition (with possible initialization). }
    procedure SkipMethodDefinition( var Pos : TPoint );
    {* Skip method definition (starting from keyword 'PROCEDURE', 'FUNCTION',
       'CONSTRUCTOR' or 'DESTRUCTOR' or from parameters list. Also all directives,
       following the definition, are skipped (like 'OVERRIDE', 'VIRTUAL', etc.) }
    procedure SkipMethodBody( var Pos : TPoint );
    {* Skip body of method (with declarations section). }
    property StartTokenPos : TPoint read FStartTokenPos;
    {* Returns starting coordinate of first character in last met token. }
    property Uses1Clause : PStrList read FUses1Entries;
    {* List of unit names in first 'USES' clause (in interface section). }
    property Uses2Clause : PStrList read FUses2Entries;
    {* List of unit names in second 'USES' clause (in implementation section). }
  end;

  function NewEntry( Owner : PRecordClass; AUnit : PPascalUnit ): PEntry;

  function NewRecordClass( Source : PPascalUnit; FStart : TPoint ): PRecordClass;

  function NewPascalUnit: PPascalUnit;

  function ComparePositions( Pos1, Pos2 : TPoint ) : Integer;
  {* Returns -1, 0 or 1 depending on what position is earlier.
     (-1, if Pos1 < Pos2, and so on). }

implementation

function ComparePositions(Pos1, Pos2: TPoint): Integer;
begin
  if Pos1.y < Pos2.y then Result := -1 else
  if Pos1.y > Pos2.y then Result := 1 else
  if Pos1.x < Pos2.x then Result := -1 else
  if Pos1.x > Pos2.x then Result := 1 else
  Result := 0;
end;

{ TRecordClass }

procedure TRecordClass.ClearEntries;
var I : Integer;
    E : PEntry;
begin
  for I := 0 to FEntries.Count - 1 do
  begin
    E := FEntries.Items[ I ];
    E.Free;
  end;
  FEntries.Clear;
  FDefStart.y := -1;
end;

procedure TRecordClass.BuildDefinition;
var E : PEntry;
    P : TPoint;
    S : String;
    Section : TSectionType;

    procedure BuildMethodDefinition;
    begin
      E := NewEntry( @Self, FUnit );
      E.FStartPos := FUnit.FStartTokenPos;
      E.FType := etMethod;
      E.FSection := Section;
      FEntries.Add( E );
      E.FName := FUnit.Scan( P );
      FUnit.SkipMethodDefinition( P );
      E.FEndPos := P;
      while P.y < FUnit.Text.Count do
      begin
        S := FUnit.Scan( P );
        if not StrIn( S, [ 'cdecl', 'register', 'dynamic', 'virtual', 'abstract',
         'export', 'external', 'far', 'forward', 'message', 'override', 'assembler',
         'overload', 'pascal', 'reintroduce', 'safecall', 'stdcall' ] ) then break;
        while (S <> ';') and (P.y < FUnit.Text.Count) do
          S := FUnit.Scan( P );
        E.FEndPos := P;
      end;
      P := E.FEndPos;
    end;

    procedure BuildPropertyDefinition;
    begin
      E := NewEntry( @Self, FUnit );
      E.FStartPos := FUnit.FStartTokenPos;
      E.FType := etProperty;
      E.FSection := Section;
      FEntries.Add( E );
      E.FName := FUnit.Scan( P );
      S := FUnit.Scan( P );
      if S = '[' then
         FUnit.SkipToSqrBracket( P );
      if S <> ';' then
         FUnit.ScanToToken( ';', P, False, False, True );
      E.FEndPos := P;
      while True do
      begin
        S := FUnit.Scan( P );
        if StrIn( S, [ 'read', 'write', 'default', 'nodefault', 'implements' ] ) then
        begin
           FUnit.ScanToToken( ';', P, False, False, True );
           E.FEndPos := P;
        end
           else
        begin
          P := E.FEndPos;
          break;
        end;
      end;
    end;

    procedure BuildFieldDefinition;
    var SS: String;
        L: PList;
        I: Integer;
    begin
      L := NewList;
      E := NewEntry( @Self, FUnit );
      E.FStartPos := FUnit.FStartTokenPos;
      E.FType := etField;
      E.FSection := Section;
      E.FName := S;
      FEntries.Add( E );
      L.Add( E );
      //FUnit.ScanToToken( ';', P, False, False, True );
      SS := FUnit.Scan( P );
      while SS = ',' do
      begin
        SS := FUnit.Scan( P );
        E := NewEntry( @Self, FUnit );
        E.FStartPos := FUnit.FStartTokenPos;
        E.FType := etField;
        E.FSection := Section;
        E.FName := SS;
        FEntries.Add( E );
        L.Add( E );
        SS := FUnit.Scan( P );
      end;
      P := FUnit.FStartTokenPos;
      FUnit.SkipDataDefinition( P );
      //E.FEndPos := P;
      for I := 0 to L.Count - 1 do
      begin
        E := L.Items[ I ];
        E.FEndPos := P;
      end;
      L.Free;
    end;

    procedure DoStartBuildDefinition;
    var Delim, S1, S0: String;
    begin
      S := FUnit.Scan( P );
      if StrEq( S, 'packed' ) then
         S := FUnit.Scan( P );
      if StrIn( S, [ 'class', 'object', 'interface' ] ) then
      begin
        S0 := S;
        S := FUnit.Scan( P );
        if S <> '(' then
        begin
          if StrIn( S, [ 'class', 'interface' ] ) then
             FAncestorName := 'TObject'
          else
             FAncestorName := '';
          if S = ';' then Exit;
          P := FUnit.FStartTokenPos;
        end
           else
        begin
          FAncestorName := FUnit.Scan( P );
          Delim := FUnit.Scan( P );
          while Delim = ',' do
          begin
            FUnit.Scan( P );
            Delim := FUnit.Scan( P );
          end;
          if Delim <> ')' then Exit;
          fDefBodyStart := P;
          S1 := FUnit.Scan( P );
          if StrEq( S0, 'interface' ) then
            if S1 = '[' then
            begin
              if not FUnit.ScanToToken( ']', P, True, True, True ) then
                Exit;
              S1 := FUnit.Scan( P );
              {if S1 <> '#' then
                P := FUnit.StartTokenPos
              else
              S1 := FUnit.Scan( P );}
            end;
          if S1 = ';' then Exit;
          P := FUnit.FStartTokenPos;
        end;
      end
         else
      if StrEq( S, 'record' ) then
      begin
        FAncestorName := '';
      end
         else Exit;
      Section := stPublic;
      while P.y <= FUnit.FText.Count - 1 do
      begin
        S := FUnit.Scan( P );
        if StrEq( S, 'end' ) then break;
        if StrIn( S, [ 'public', 'published', 'automated' ] ) then
        begin
          Section := stPublic;
          continue;
        end;
        if StrEq( S, 'private' ) then
        begin
          Section := stPrivate;
          continue;
        end;
        if StrEq( S, 'protected' ) then
        begin
          Section := stProtected;
          continue;
        end;
        //P0 := P;
        if StrEq( S, 'class' ) then
          S := FUnit.Scan( P );
        if StrIn( S, [ 'function', 'procedure', 'constructor', 'destructor' ] ) then
        begin
          BuildMethodDefinition;
          continue;
        end;
        if StrEq( S, 'property' ) then
        begin
          BuildPropertyDefinition;
          continue;
        end;
        BuildFieldDefinition;
      end;
      FUnit.Scan( P ); // skip ';'
    end;

begin
  if FEntries.Count > 0 then Exit;
  P := FDefStart;
  FUnit.Scan( P ); P := FUnit.FStartTokenPos;
  FDefStart := P;
  fDefBodyStart := P;
  FDefEnd := FDefStart;
  if P.y < 0 then Exit;
  S := FUnit.Scan( P );
  if S = '' then Exit;
  FName := S;
  if FUnit.Scan( P ) <> '=' then Exit;
  DoStartBuildDefinition;
  FDefEnd := P;
end;

function TRecordClass.FindEntry(EntryTypes: TEntryTypeSet;
  Sections: TSectionTypeSet; const AName: String): PEntry;
var I : Integer;
    E : PEntry;
begin
  Result := nil;
  for I := 0 to FEntries.Count - 1 do
  begin
    E := Entries[ I ];
    if (EntryTypes = []) or (E.EntryType in EntryTypes) then
    if (Sections = []) or (E.Section in Sections) then
    if StrEq( E.Name, AName ) then
    begin
      Result := Entries[ I ];
      Exit;
    end;
  end;
end;

procedure TRecordClass.SearchMethodImplementations;
var P, P0, BB, EB : TPoint;
    S, FT : String;
    I : Integer;
    E : PEntry;
    EntryFound : Boolean;
begin
  P := MakePoint( 0, 0 );
  if FUnit.FImplementationPos.y < 0 then
     if not FUnit.ScanToToken( 'implementation', P, False, False, True ) then
        Exit;
  P := FUnit.FImplementationPos;
  FT := FName;
  E := nil;
  while P.y <= FUnit.FText.Count - 1 do
  begin
    S := FUnit.Scan( P );
    P0 := FUnit.FStartTokenPos;
    if StrIn( S, [ 'procedure', 'function', 'constructor', 'destructor' ] ) then
    begin
      if StrEq( FUnit.Scan( P ), FT ) then
      if FUnit.Scan( P ) = '.' then
      begin
        S := FUnit.Scan( P );
        EntryFound := False;
        for I := 0 to FEntries.Count - 1 do
        begin
          E := Entries[ I ];
          if (E.EntryType = etMethod) and
             StrEq( E.Name, S ) then
          begin
            EntryFound := True;
            break;
          end;
        end;
        if EntryFound then
        begin
          if not FUnit.ScanToToken( 'begin', P, False, True, False ) then
             continue;
          BB := P;
          P := FUnit.FStartTokenPos;
          FUnit.SkipMethodBody( P );
          P := FUnit.FStartTokenPos;
          S := FUnit.Scan( P ); // 'end'
          EB := FUnit.FStartTokenPos;
          S := FUnit.Scan( P );
          if S = ';' then
          begin
            E.FBeginImplement := P0;
            E.FBeginBody := BB;
            E.FEndBody := EB;
          end;
        end;
      end;
    end;
  end;
end;

function NewRecordClass(Source: PPascalUnit; FStart: TPoint): PRecordClass;
begin
  New( Result, Create );
  Result.FUnit := Source;
  Result.FDefStart := FStart;
  Result.FEntries := NewList;
  //-- scan here declaration of fields, methods and properties:
  Result.BuildDefinition;
end;

destructor TRecordClass.Destroy;
begin
  ClearEntries;
  FEntries.Free;
  inherited;
end;

function TRecordClass.GetEntries(Idx: Integer): PEntry;
begin
  Result := FEntries.Items[ Idx ];
end;

function TRecordClass.GetEntryCount: Integer;
begin
  Result := FEntries.Count;
end;

{ TPascalUnit }

procedure TPascalUnit.BuildDefinitions;
var P, P0, Pos : TPoint;
    S : String;
  procedure BuildTypeDefinitions;
  var Name : String;
      RC : PRecordClass;
      E : PEntry;
      S0 : String;
  begin
    while P.y < FText.Count do
    begin
      Pos := P;
      Name := Scan( P );
      if Scan( P ) <> '=' then
      begin
        P := Pos;
        SkipTypeDefinition( P );
      end
         else
      begin
        S := Scan( P );
        if StrEq( S, 'packed' ) then
           S := Scan( P );
        if StrIn( S, [ 'class', 'record', 'object', 'interface' ] ) then
        begin
          if StrIn( S, [ 'class', 'object', 'interface' ] ) then
          begin
            P0 := P;
            S0 := Scan( P );
            if StrEq( S, 'object' ) and (S0 <> '(') then
            begin
              P := P0;
              RC := NewRecordClass( @Self, Pos );
              P := RC.FDefEnd;
              FDecls.Add( RC );
            end
               else
            if StrEq( S, 'interface' ) and (S0 = '(') then
            begin
              ScanToToken( ')', P, True, True, True );
              RC := NewRecordClass( @Self, Pos );
              P := RC.FDefEnd;
              FDecls.Add( RC );
            end
               else
            if (S0 = '(') and (Length( Scan( P ) ) >= 0)
            and (Scan( P ) = ')') or (S0 <> ';') then
            begin
              RC := NewRecordClass( @Self, Pos );
              P := RC.FDefEnd;
              FDecls.Add( RC );
            end;
          end
             else
          begin
            P0 := P;
            S0 := Scan( P );
            if S0 <> ';' then
            begin
              P := P0;
              RC := NewRecordClass( @Self, Pos );
              P := RC.FDefEnd;
              FDecls.Add( RC );
            end;
          end;
        end
           else
        begin
          E := NewEntry( nil, @Self );
          E.FName := Name;
          E.FType := etType;
          E.FSection := stPublic;
          E.FStartPos := Pos;
          P := Pos;
          SkipTypeDefinition( P );
          E.FEndPos := P;
          FDecls.Add( E );
        end;
      end;
      Pos := P;
      S := Scan( P );
      if StrIn( S, [ 'type', 'const', 'var', 'threadvar', 'implementation',
                'procedure', 'function' ] ) then
      begin
        P := Pos;
        break;
      end;
      P := Pos;
    end;
  end;

  procedure BuildDataDefinitions;
  var E : PEntry;
  begin
    while P.y < FText.Count do
    begin
      E := NewEntry( nil, @Self );
      Pos := P;
      E.FName := Scan( P );
      P := Pos;
      E.FType := etField;
      E.FSection := stPublic;
      E.FStartPos := P;
      SkipDataDefinition( P );
      E.FEndPos := P;
      FDecls.Add( E );

      Pos := P;
      S := Scan( P );
      if StrIn( S, [ 'type', 'const', 'var', 'threadvar', 'implementation',
                'procedure', 'function' ] ) then
      begin
        P := Pos;
        break;
      end;
      P := Pos;
    end;
  end;

  procedure BuildProcFunDefinition;
  var E : PEntry;
  begin
    S := Scan( P );
    P := Pos; Scan( P ); Pos := FStartTokenPos; P := Pos;
    E := NewEntry( nil, @Self );
    E.FName := S;
    E.FType := etMethod;
    E.FSection := stPublic;
    E.FStartPos := Pos;
    SkipMethodDefinition( P );
    E.FEndPos := P;
    FDecls.Add( E );
  end;

begin
  if FText.Count = 0 then Exit;
  P := MakePoint( 0, 0 );
  if not StrEq( Scan( P ), 'unit' ) then Exit;
  FUnitName := Scan( P );
  if Scan( P ) <> ';' then Exit;
  if not StrEq( Scan( P ), 'interface' ) then Exit;
  if not StrEq( Scan( P ), 'uses' ) then Exit;
  ScanUsesEntries( FUses1Entries, P );
  while P.y < FText.Count do
  begin
    Pos := P;
    S := Scan( P );
    if StrEq( S, 'type' ) then
       BuildTypeDefinitions
    else
    if StrIn( S, [ 'const', 'var', 'threadvar' ] ) then
       BuildDataDefinitions
    else
    if StrIn( S, [ 'function', 'procedure' ] ) then
       BuildProcFunDefinition
    else
       break;
  end;
  FImplementationPos := P;
end;

procedure TPascalUnit.ScanUsesEntries( WhatUses : PStrList; var Pos : TPoint );
var P : TPoint;
    S : String;
begin
  if WhatUses.Count > 0 then Exit;
  P := Pos;
  if P.y < 0 then Exit;
  while P.y <= FText.Count - 1 do
  begin
    S := Scan( P );
    WhatUses.Add( S );
    S := Scan( P );
    if StrEq( S, 'in' ) then
    begin
      while P.y <= FText.Count - 1 do
      begin
        S := Scan( P );
        if (S = ',') or (S = ';') then break;
      end;
    end;
    if S <> ',' then break;
  end;
  Pos := P;
end;

procedure TPascalUnit.Clear;
begin
  FText.Clear;
  FCleanText.Clear;
  FUsesClause1Pos.y := -1;
  FUsesClause1End.y := -1;
  FImplementationPos.y := -1;
  FUsesClause2Pos.y := -1;
  ClearDeclsAndGlobals;
  FUses1Entries.Clear;
  FUses2Entries.Clear;
end;

function NewPascalUnit: PPascalUnit;
begin
  New( Result, Create );
  Result.FText := NewStrList;
  Result.FCleanText := NewStrList;
  Result.FDecls := NewList;
  Result.FUses1Entries := NewStrList;
  Result.FUses2Entries := NewStrList;
  Result.Clear;
end;

destructor TPascalUnit.Destroy;
begin
  Clear;
  FUnitPath := '';
  FText.Free;
  FCleanText.Free;
  FDecls.Free;
  FUses2Entries.Free;
  inherited;
end;

function TPascalUnit.GetCleanText(Idx: Integer): String;
label search_string, search_comments;
var I, J, K, K1, L, L1, M : Integer;
    S : String;

    procedure SubstituteString( var S : String; M : Integer );
    var I : Integer;
    begin
      S[ M ] := '#';
      for I := M + 1 to Length( S ) do
      begin
        if S[ I ] = '''' then
           if ( I = Length( S ) ) or ( S[ I + 1 ] <> '''' ) then
           begin
             S[ I ] := '0';
             break;
           end
              else
           begin
             S[ I ] := '0';
             S[ I + 1 ] := '0';
           end
        else
          S[ I ] := '0';
      end;
    end;

begin
  while Idx >= FCleanText.Count do
  begin
    if FCleanText.Count = FText.Count then break;
    I := FCleanText.Count - 1;
    while I < FText.Count - 1 do
    begin
      if I > Idx then break;
      Inc( I );
      S := FText.Items[ I ];
      if S = '' then
      begin
        FCleanText.Add( '' );
        continue;
      end;
    search_comments:
      J := IndexOfStr( S, '//' );
      K := IndexOfChar( S, '{' );
      L := IndexOfStr( S, '(*' );
    search_string:
      M := IndexOfChar( S, '''' );
      if (J < 0) and (K < 0) and (L < 0) and (M < 0) then
      begin
        FCleanText.Add( TrimRight( S ) );
        continue;
      end;
      if M > 0 then
      begin
        if ((J > 0) and (M < J) or (J <= 0)) and
           ((K > 0) and (M < K) or (K <= 0)) and
           ((L > 0) and (M < L) or (L <= 0)) then
        begin
          SubstituteString( S, M );
          goto search_comments;
        end;
      end;
      if J > 0 then
      begin
        if ((J < K) or (K <= 0)) and ((J < L) or (L <= 0)) then
        begin
          S := Copy( S, 1, J - 1 );
          FCleanText.Add( TrimRight( S ) );
          continue;
        end;
      end;
      if (K > 0) and ((L <= 0) or (K < L)) then
      begin
         K1 := IndexOfChar( S, '}' );
         if (K1 > K) then
         begin
           for K := K to K1 do
               S[ K ] := ' ';
           goto search_comments;
         end;
         S := Copy( S, 1, K - 1 );
         while True do
         begin
           FCleanText.Add( TrimRight( S ) );
           Inc( I );
           if I >= FText.Count then break;
           S := FText.Items[ I ];
           K1 := IndexOfChar( S, '}' );
           if K1 > 0 then
           begin
             for K := 1 to K1 do
                 S[ K ] := ' ';
             goto search_comments;
           end;
           S := '';
         end;
      end;
      if (L > 0) and ((K <= 0) or (L < K)) then
      begin
         S[ L + 1 ] := ' ';
         L1 := IndexOfStr( S, '*)' );
         if (L1 > L) then
         begin
           Inc( L1 );
           for L := L to L1 do
               S[ L ] := ' ';
           goto search_comments;
         end;
         S := Copy( S, 1, L - 1 );
         while True do
         begin
           FCleanText.Add( TrimRight( S ) );
           Inc( I );
           if I >= FText.Count then break;
           S := FText.Items[ I ];
           L1 := IndexOfStr( S, '*)' );
           if L1 > 0 then
           begin
             Inc( L1 );
             for L := 1 to L1 do
                 S[ L ] := ' ';
             goto search_comments;
           end;
           S := '';
         end;
      end;
      FCleanText.Add( TrimRight( S ) );
    end;
  end;
  Result := FCleanText.Items[ Idx ];
end;

function TPascalUnit.GetDeclsCount: Integer;
begin
  Result := FDecls.Count;
end;

function TPascalUnit.GetDecls(Idx: Integer): PObj;
begin
  Result := FDecls.Items[ Idx ];
end;

procedure TPascalUnit.Load;
var Buf : String;
begin
  if FText.Count <> 0 then Exit;
  if FUnitPath = '' then Exit;
  if Buf = '' then
  begin
    if FileExists( FUnitPath ) then
    begin
      Clear;
      FText.LoadFromFile( FUnitPath );
    end;
  end;
end;

function TPascalUnit.Scan(var Pos: TPoint): String;
label loop_end;
var S : String;
    P : TPoint;
    I : Integer;
begin
  while Pos.y <= FText.Count - 1 do
  begin
    if Pos.x >= Length( CleanText[ Pos.y ] ) then
    begin
      Pos.x := 0;
      Inc( Pos.y );
      continue;
    end;
    S := CleanText[ Pos. y ];
    while S[ Pos.x + 1 ] <= ' ' do
    begin
      Inc( Pos.x );
      if Pos.x >= Length( S ) then
      begin
        Pos.x := 0;
        Inc( Pos.y );
        goto loop_end;
      end;
    end;
    FStartTokenPos := Pos;
    if S[ Pos.x + 1 ] in [ 'A'..'Z', 'a'..'z', '_', ''..'', ''..'', '', '' ] then
    begin
      Result := '';
      while Pos.x < Length( S ) do
      begin
        if S[ Pos.x + 1 ] in [ 'A'..'Z', 'a'..'z', '_', '0'..'9', ''..'', ''..'', '', '' ] then
           Result := Result + S[ Pos.x + 1 ]
        else
           break;
        Inc( Pos.x );
      end;
      Exit;
    end;
    if S[ Pos.x + 1 ] in [ '+', '-', '#', '$', '0'..'9' ] then
    begin
      P := Pos;
      Result := '';
      while Pos.x < Length( S ) do
      begin
        if S[ Pos.x + 1 ] in [ '#', '$', '+', '-', '0'..'9', 'A'..'F', 'a'..'f' ] then
           Result := Result + S[ Pos.x + 1 ]
        else
           break;
        Inc( Pos.x );
      end;

      if Length( Result ) > 1 then
      if Result[ 1 ] = '#' then
      begin
        for I := 2 to Length( Result ) do
            if Result[ I ] <> '0' then Exit;
        S := Text.Items[ P.y ];
        Result := Copy( S, P.x + 1, Length( Result ) );
      end;

      Exit;
    end;
    Result := S[ Pos.x + 1 ];
    Inc( Pos.x );
    Exit;
  loop_end:
  end;
end;

function TPascalUnit.ScanToToken(const Token: String;
  var Pos: TPoint; BeforeImplementation, IgnoreData, NoMethods : Boolean): Boolean;
var S : String;
begin
  Result := False;
  while Pos.y <= FText.Count - 1 do
  begin
    S := Scan( Pos );
    if (FImplementationPos.y < 0) and StrEq( S, 'implementation' ) then
    begin
      FImplementationPos := Pos;
      if BeforeImplementation then
         Exit;
    end
       else
    if (FUsesClause2Pos.y < 0) and (Pos.y > FImplementationPos.y) and
       (FimplementationPos.y >= 0) and StrEq( S, 'uses' ) then
      FUsesClause2Pos := Pos
    else
    if (FUsesClause1Pos.y < 0) and ((FImplementationPos.y < 0) or
       (Pos.y < FImplementationPos.y)) and StrEq( S, 'uses' ) then
      FUsesClause1Pos := Pos;

    if StrEq( S, Token ) then
    begin
      Result := True;
      Exit;
    end;
    if (not IgnoreData) and StrIn( S, [ 'const', 'var', 'threadvar' ] ) then
       SkipDataDefinitions( Pos )
    else
    if StrEq( S, 'type' ) then
       SkipTypeDefinitions( Pos )
    else
    if not NoMethods and StrIn( S, [ 'procedure', 'function', 'constructor', 'destructor' ] ) then
    begin
       SkipMethodDefinition( Pos );
       if FImplementationPos.y > 0 then
       if ComparePositions( Pos, FImplementationPos ) > 0 then
          SkipMethodBody( Pos );
    end;
  end;
end;

procedure TPascalUnit.SetUnitPath(const Value: String);
begin
  if FUnitPath = Value then Exit;
  Clear;
  FUnitPath := Value;
  Load;
end;

function TPascalUnit.SkipDataDefinition(var Pos: TPoint) : Boolean;
var S : String;
    P0 : TPoint;
    FirstChar: Char;
begin
  while Pos.y <= FText.Count - 1 do
  begin
        S := Scan( Pos );
        if StrIn( S, ['=', ':'] ) then
        begin
          FirstChar := S[ 1 ];
          while Pos.y <= FText.Count - 1 do
          begin
            S := Scan( Pos );
            if S = '(' then
              SkipToBracket( Pos );
            if S = ';' then
              break;
            if FirstChar = ':' then
            if StrEq( S, 'record' ) then
              ScanToEnd( Pos );
          end;
        end;
        if S = ';' then
           break;
  end;
  Result := True;
  P0 := Pos;
  if Pos.y <= FText.Count - 1 then
        begin
          S := Scan( Pos );
          if StrIn( S, [ 'type', 'uses', 'implementation',
                         'procedure', 'function', 'constructor', 'destructor',
                         'var', 'const', 'threadvar' ] ) then
            Result := False;
          Pos := P0;
        end;
end;

procedure TPascalUnit.SkipDataDefinitions(var Pos: TPoint);
begin
  while (Pos.y <= FText.Count - 1)
        and SkipDataDefinition( Pos ) do;
end;

procedure TPascalUnit.SkipMethodBody(var Pos: TPoint);
var S : String;
  procedure SkipBeginToEnd;
  var S : String;
  begin
    while Pos.y <= FText.Count - 1 do
    begin
      S := Scan( Pos );
      if StrEq( S, 'end' ) then Exit;
      if StrIn( S, [ 'begin', 'case', 'try' ] ) then
         SkipBeginToEnd;
    end;
  end;
begin
  while Pos.y <= FText.Count - 1 do
  begin
    S := Scan( Pos );
    if StrIn( S, [ 'var', 'const', 'threadvar' ] ) then
       SkipDataDefinitions( Pos )
    else
    if StrEq( S, 'type' ) then
       SkipTypeDefinitions( Pos )
    else
    if StrIn( S, [ 'procedure', 'function', 'constructor', 'destructor' ] ) then
    begin
       SkipMethodDefinition( Pos );
       SkipMethodBody( Pos );
    end
       else
    if StrIn( S, [ 'asm', 'begin' ] ) then
    begin
       SkipBeginToEnd;
       break;
    end;
  end;
end;

procedure TPascalUnit.SkipMethodDefinition(var Pos: TPoint);
var S : String;
    P : TPoint;
begin
  while Pos.y <= FText.Count - 1 do
  begin
    S := Scan( Pos );
    if S = '' then break
    else
    if S = ';' then break
    else
    if S = '(' then
       SkipToBracket( Pos )
    else
    if S = '[' then
       SkipToSqrBracket( Pos );
  end;
  while Pos.y < FText.Count do
  begin
    P := Pos;
    S := Scan( Pos );
    if not StrIn( S, [ 'cdecl', 'register', 'dynamic', 'virtual', 'abstract',
         'export', 'external', 'far', 'forward', 'message', 'override',
         'assembler', 'overload', 'pascal', 'reintroduce', 'safecall',
         'stdcall' ] ) then
    begin
      Pos := P;
      break;
    end;
    while (S <> ';') and (Pos.y < FText.Count) do
       S := Scan( Pos );
  end;
end;

procedure TPascalUnit.SkipToBracket(var Pos: TPoint);
var S : String;
begin
  while Pos.y <= FText.Count - 1 do
  begin
    S := Scan( Pos );
    if S = ')' then Exit;
    if S = '(' then SkipToBracket( Pos );
  end;
end;

procedure TPascalUnit.SkipToSqrBracket(var Pos: TPoint);
var S : String;
begin
  while Pos.y <= FText.Count - 1 do
  begin
    S := Scan( Pos );
    if S = ']' then Exit;
    if S = '[' then SkipToSqrBracket( Pos );
  end;
end;

procedure TPascalUnit.SkipTypeDefinition(var Pos: TPoint);
var BeforeImplementation : Boolean;
var S : String;
begin
  BeforeImplementation := (FImplementationPos.y < 0)
                       or (ComparePositions( Pos, FImplementationPos ) < 0);
  if ScanToToken( '=', Pos, BeforeImplementation, False, True ) then
  begin
    S := Scan( Pos );
    if StrIn( S, [ 'procedure', 'function' ] ) then
    begin
      SkipMethodDefinition( Pos );
      Exit;
    end;
    if StrEq( S, 'packed' ) then
       S := Scan( Pos );
    if StrIn( S, [ 'class', 'object' ] ) then
    begin
      if StrEq( Scan( Pos ), 'of' ) then
      begin
         ScanToToken( ';', Pos, BeforeImplementation, False, True );
      end
         else
      if S = ';' then
      else
      begin
        ScanToEnd( Pos );
        //ScanToToken( 'end', Pos, BeforeImplementation, False, True );
        Scan( Pos ); // skip ';'
      end;
    end
       else
    if StrIn( S, [ 'record', 'interface' ] ) then
    begin
       ScanToEnd( Pos );
       //ScanToToken( 'end', Pos, BeforeImplementation, False, True );
       Scan( Pos ); // Skip ';'
    end
    else
       ScanToToken( ';', Pos, BeforeImplementation, False, True );
  end
     else Exit;
end;

procedure TPascalUnit.SkipTypeDefinitions(var Pos: TPoint);
var S : String;
begin    
  while Pos.y <= FText.Count - 1 do    
  begin    
    SkipTypeDefinition( Pos );    
    S := Scan( Pos );    
    if S = '' then break;    
    if S = ';' then    
       S := Scan( Pos );    
    if StrIn( S, [ 'uses', 'const', 'var', 'threadvar', 'procedure', 'function',
              'constructor', 'destructor' ] ) then
    begin    
      Pos := FStartTokenPos;    
      break;    
    end;
  end;    
end;

procedure TPascalUnit.ClearDeclsAndGlobals;
var I : Integer;
begin
  for I := FDecls.Count - 1 downto 0 do
    PObj( FDecls.Items[ I ] ).Free;
  FDecls.Clear;
end;

function TPascalUnit.ScanToEnd(var Pos: TPoint): Boolean;
var Lev: Integer;
    S: String;
begin
  Lev := 1;
  Result := TRUE;
  while Lev > 0 do
  begin
    S := Scan( Pos );
    if StrEq( S, 'end' ) then
      Dec( Lev )
    else
    if StrIn( S, [ 'begin', 'record', 'object', 'class' ] ) then
      Inc( Lev );
    if S = '' then
    begin
      Result := FALSE;
      break;
    end;
  end;
end;

{ TEntry }

function NewEntry( Owner: PRecordClass; AUnit : PPascalUnit ): PEntry;
begin
  New( Result, Create );
  Result.FIsEntry := True;
  Result.FOwner := Owner;
  Result.FUnit := AUnit;
end;

function TEntry.GetBeginBody: TPoint;
begin
  Result := FBeginBody;
  if FType <> etMethod then Exit;
  if FOwner = nil then Exit;
  if FEndBody.y = 0 then
     FOwner.SearchMethodImplementations;
  Result := FBeginBody;
  if FEndBody.y = 0 then
     Result := FEndBody;
end;

function TEntry.GetEndBody: TPoint;
begin
  Result := FEndBody;
  if FType <> etMethod then Exit;
  if FUnit = nil then Exit;
  if FEndBody.y = 0 then
     FOwner.SearchMethodImplementations;
  Result := FEndBody;
end;

function TEntry.IsBodyEmpty: Boolean;
var BB, EB : TPoint;
    S : String;
    Y, I1, I2 : Integer;
begin
  BB := BeginBody;
  EB := EndBody;
  Result := False;
  if BB.y <= 0 then Exit;
  if EB.y <= 0 then Exit;
  Result := True;
  for Y := BB.y to EB.y do
  begin
    S := FUnit.FText.Items[ Y ];
    I1 := 1; I2 := Length( S );
    if Y = BB.y then
       I1 := BB.x + 1;
    if Y = EB.y then
       I2 := EB.x;
    if Trim( Copy( S, I1, I2 - I1 ) ) <> '' then
    begin
      Result := False;
      break;
    end;
  end;
end;

end.
